function [u prelag Tu dev_dssd dev_Tu dev_u] = doElasticReg(R,T,mu,lambda,numiters,varargin)
    
    prereg        = getVar(varargin,'prereg',[]);
    donormalize   = getVar(varargin,'normalize',true);
    donormalize2  = getVar(varargin,'normalize2',false);
    EXTWINSIZE    = getVar(varargin,'EXTWINSIZE',3);
    maxdeform     = getVar(varargin,'maxdeform',EXTWINSIZE);
    fixedidcs     = getVar(varargin,'fixedidcs',[]);
        
    if isempty(R) || isempty(T)
        error('R and/or T is empty.');
    end

    isvectorR = isvector(R);
    if isvectorR
        R = R(:)';
        T = T(:)';
        spacing = 1;
    else
        spacing=ones(1,ndims(R)); % grid spacing
    end    
    
    T_orig = T;
    % normalize data
    if donormalize
        R = (R-repmat(mean(R,2),1,size(R,2)))./repmat(std(R,[],2),1,size(R,2));
        T = (T-repmat(mean(T,2),1,size(T,2)))./repmat(std(T,[],2),1,size(T,2));
    end
    
    if donormalize2
        R = R./repmat(sum(R,2),1,size(R,2))*255;
        T = T./repmat(sum(T,2),1,size(T,2))*255;
    end
     
    % pre-registration ----------------------------------------------------------------
    if ~isempty(prereg)

        dssd = zeros(1,length(prereg));
        cands = zeros(size(T,1),size(T,2),length(prereg));
%         ss = speechsignal();
        for k=1:length(prereg)
%             cands(:,:,k) = getFeature(ss,'IDENT','usefeatsignal',T,'vtln','welling','warptype','add', ...
%                                   'warplinreg',true,'warpfactor',prereg(k));
            cands(:,:,k) = translateAlongFreqAxis(T,prereg(k));
            dssd(k) = sum(sum((cands(:,:,k)-R).^2));
        end
        [~,minidcs] = min(dssd,[],2);
        T = cands(:,:,minidcs);        
        
        prelag = prereg(minidcs);
        
    else
        prelag = 0;
    end
    % =================================================================================       
    
    nop = size(R,2);
    numclass = size(R,1);
    numelR = numel(R);
    u_l = zeros(numelR,length(spacing));  % deformation field
    
    % extend T with linear slopes at both ends ----------------------------------------       
    T_extd = extendSides(T,EXTWINSIZE);
    
    % mesh grids for later interpolation calls
    [x1ext x2ext]   = ndgrid((-EXTWINSIZE+1):nop+EXTWINSIZE,1:numclass);
    [xi1ext xi2ext] = ndgrid(1:nop,1:numclass); %#ok<ASGLU>
    % =================================================================================
                
    % build coefficient matrix L ------------------------------------------------------
    if ~exist('elastregcoeffmatrix','dir')
        mkdir('elastregcoeffmatrix');
    end
    
    matrixfilename = 'elastregcoeffmatrix';
    for i=1:length(spacing)
        matrixfilename = [matrixfilename '-' int2str(size(R,i))];
    end
    
    matrixfilename = ['elastregcoeffmatrices/' matrixfilename '-' num2str(lambda) '-' num2str(mu) '.mat'];
    
    if exist(matrixfilename,'file')
        disp('Loading coefficient matrix.');
        load(matrixfilename);
    else
        L = generateAMatrix(size(R),lambda,mu,spacing);
        save(matrixfilename,'L');
    end
    % =================================================================================
       
    dev_dssd = zeros(numiters,1); % development of distance measure
    
    if lambda~=-mu
        % ensure positive definiteness in case of elastic regularizer
        calc_k = min(min(spacing), prod(spacing)/(lambda+mu) );
    else
        calc_k = min(spacing);
    end    
        
    % save threshold for k, consider positive definiteness of L
    k_thresh = calc_k;    
    k_step = calc_k/100;
    
    THRESH = 1e-2;
    ctr = 1;    % counter for stopping after maximimum number of iterations
    Tu = T;
    
    startDist = sum((R(:)-T(:)).^2);
    
    disp('Starting iterative solution process.');
    while ctr <= numiters % use a fixed number of iterations as stopping criterion

        % calculate f        
        [fx2 fx1] = gradient(Tu,spacing(2),spacing(1));
        
        % n-dimensional force term
        fu1 = (R - Tu) .* fx1;
        fu2 = (R - Tu) .* fx2;
        
        fu1(:) = 0; % only allow dispacements along frequency axis

        tu1 = fu1(:) * calc_k + u_l(:,1);
        tu2 = fu2(:) * calc_k + u_l(:,2);

        L2 = speye(length(spacing)*numelR) - L * calc_k;

        u_new = solveCG( L2,[tu1; tu2]);
        
        u_new = reshape(u_new,size(u_l,1),length(spacing));
        
        u_new(:,1) = 0;
        tempidcs = abs(u_new(:,2))>maxdeform;
        u_new(tempidcs,2) = sign(u_new(tempidcs,2))*maxdeform;
        
        if ~isempty(fixedidcs)
            tempt = reshape( u_new(:,2),size(T));
            tempt(fixedidcs==1,:) = 0;
            u_new(:,2) = tempt(:);
        end
        
        Tu_old = Tu;                  
        Tu = interpTu_ext( u_new, T_extd, EXTWINSIZE, x1ext,x2ext,xi2ext);                                 
   
        curDist = sum((R(:)-Tu(:)).^2);        
        oldDist = sum((R(:)-Tu_old(:)).^2);        

        tempd = curDist/startDist;
        if tempd<THRESH % && iter==-1
            break;
        end
        if curDist>=1.01*oldDist
            calc_k = calc_k/2;
        elseif calc_k<k_thresh  % while calc_k is smaller than its inital value
            calc_k = calc_k+k_step;
        end

        u_l = u_new;
        dev_dssd(ctr) = tempd;

        if ctr>100
            polycoeffs = polyfit(1:100,dev_dssd(ctr-99:ctr)',1);
            if abs(polycoeffs(1))<=1e-4
                break;
            end
        end        
        
        if mod(ctr,100)==0
            fprintf('.');   
        end
        if mod(ctr,1000)==0
            fprintf('\n');
        end

        ctr = ctr+1; % increment counter
    end % end while    
    fprintf('\n');
            
    if nargout>=3
        Tu = interpTu( u_l, T_orig );     
    end
    
    if nargout>=4
        dev_dssd = dev_dssd(1:ctr-1);
    end
    
    %     % finalize displacement fields
    u_l(:,2) = u_l(:,2) - prelag;
    u = u_l;   
    
end

% ==========================================
% supplementary functions ==================
% ==========================================

% extend T with linear slopes at both ends 
function T_extd = extendSides(T,EXTWINSIZE)
    CONSIDCS = 3;
    meanslope = mean(T(:,end-CONSIDCS+1:end)-T(:,end-CONSIDCS:end-1),2);
    extvals = repmat(T(:,end),1,EXTWINSIZE) + ...
              repmat(meanslope,1,EXTWINSIZE).*((1:EXTWINSIZE)'*ones(1,size(T,1)))';
    extvals = max(min(T(:)),extvals);
    T_extd = [T extvals];

    meanslope = mean(T(:,1:CONSIDCS)-T(:,2:CONSIDCS+1),2);
    extvals = repmat(T(:,1),1,EXTWINSIZE) + ...
              repmat(meanslope,1,EXTWINSIZE).*((EXTWINSIZE:-1:1)'*ones(1,size(T,1)))';
    extvals = max(min(T(:)),extvals);
    T_extd = [extvals T_extd];        
end


% and colcount columns
function Tu = interpTu( u_l, T )

    [x1,x2] = ndgrid(1:size(T,1),1:size(T,2));    
    xi1 = x1 + reshape( u_l(:,1),size(T));
    xi2 = x2 + reshape( u_l(:,2),size(T));      
    xi1(xi1<1) = 1;
    xi1(xi1>size(xi1,1)) = size(xi1,1);
    xi2(xi2<1) = 1;
    xi2(xi2>size(xi2,2)) = size(xi2,2);
       
    Tu = interpn(x1,x2,T,xi1,xi2,'*linear',0);
    
    return
end

function Tu = interpTu_ext( u_l, T, EXTWINSIZE, ...
                        x1ext,x2ext,xi2ext)

    xi1ext = x1ext((EXTWINSIZE+1):(end-EXTWINSIZE),:);
    xi1ext = xi1ext + reshape( u_l(:,2),size(T,1),size(T,2)-2*EXTWINSIZE)';       
    Tu = interpn(x2ext',x1ext',T,xi2ext',xi1ext','*linear',0);
    
    return
end

function output = translateAlongFreqAxis(input,warpfactor)

    % use linear regression for boundary extrapolation
    winsize = max(4,abs(warpfactor));
    if warpfactor>=0
        xvec  = 1:winsize;
        mean_x= mean(xvec);
        mean_y= mean(input(:,xvec),2);
        y_diff = (input(:,xvec)-repmat(mean_y,1,winsize));
        x_diff = xvec-repmat(mean_x,1,winsize);
        b = sum(repmat(x_diff,size(input,1),1) .* y_diff,2) ./sum(x_diff.^2,2);
        a = mean_y-b.*mean_x;

        % extend input matrix by extrapolated values
        extrapvals = (-winsize+1):0;
    %                     output = [max(1e-7,repmat(a,1,winsize)+repmat(b,1,winsize).*repmat(extrapvals,size(output,1),1)) output];
        input = [max(min(input(:)),repmat(a,1,winsize)+repmat(b,1,winsize).*repmat(extrapvals,size(input,1),1)) input];
        xorig = (-winsize+1):(size(input,2)-winsize);
    elseif warpfactor<0            
        xvec  = (size(input,2)-winsize+1):size(input,2);
        mean_x= mean(xvec);
        mean_y= mean(input(:,xvec),2);
        y_diff = (input(:,xvec)-repmat(mean_y,1,winsize));
        x_diff = xvec-repmat(mean_x,1,winsize);
        b = sum(repmat(x_diff,size(input,1),1) .* y_diff,2) ./sum(x_diff.^2,2);
        a = mean_y-b.*mean_x;

        % extend input matrix by extrapolated values
        extrapvals = (size(input,2)+1):(size(input,2)+winsize);
    %                     output = [output max(1e-7,repmat(a,1,winsize)+repmat(b,1,winsize).*repmat(extrapvals,size(output,1),1))];
        input = [input max(min(input(:)),repmat(a,1,winsize)+repmat(b,1,winsize).*repmat(extrapvals,size(input,1),1))];
        xorig = 1:size(input,2);
    end
    [X1 X2]   = meshgrid(xorig,1:size(input,1)); % original grid            
    indices = xorig-warpfactor; % new grid
    % check boundaries
    indices(indices<xorig(1))=xorig(1);
    indices(indices>xorig(end))=xorig(end);
    [X1i X2i] = meshgrid(indices,1:size(input,1));    
    output = interp2(X1,X2,input,X1i,X2i,'linear'); % interpolate on new grid
    if warpfactor>=0
        output = output(:,(winsize+1):end);
    else
        output = output(:,1:end-winsize);
    end    
end

% Returns a variable from an variable argument cell, e.g. if varargin
% contains the cells {'var1',123}, getVar(varargin,'var1') will return 123.
%
% INPUT
% args - cell array, usually 'varargin'
% name - name of the argument to be found, string
% default - default value for the parameter. Is set if the name of the
%           parameter is not found.
%
% OUTPUT
% r - value of the parameter
%
% AUTHOR
% Florian Mueller, 2007
%
function r = getVar( args, name, default )
    r = default;
    for i=1:length(args)
        if iscell( args{i} )
            r = getVar( args{i},name,default );
            return
        end
        if strcmp(args{i},name)==1
            r=args{i+1};
            break;
        end
    end
    
    return
end


