% Solves a linear system of equations with A symmetric and positive
% definit with the conjugated gradients method.
%
% INPUT:
% A - Coefficient matrix A
% b - Vector of the right side of the LSE Ax=b
% varargin - Optional. An epsilon as a stop criterion. Set to 1e-4 by
%            default.
%
% OUTPUT:
% res - Solution x of the LSE Ax=b
%
function  res = solveCG( A, b, varargin )
    
    if nargin>2
        epsi = varargin{1};
    else
        epsi = 1e-5;
    end;
    
    x = zeros(size(A,1),1); % result vector
    g = A * x - b;          % gradient
    d = -g;                 % direction
    k = 1;
    sx = size(A,1);
    
    while norm(g,2)>epsi && k<=sx   % while norm of gradient>epsilon repeat ...

        tn = norm(g,2)^2;   
        Ad = A*d;
        
        t  = tn/(d'*Ad);   % step size
        x  = x + t*d;      % update position
        g  = g + t*Ad;     % update gradient
        beta = norm(g,2)^2/tn;
        d  = -g + beta*d;   % update direction
        k = k+1;
     end
    
    res = x;
    
    return
end